import configparser
import json
import logging
import os
from collections import OrderedDict
from pathlib import Path

import yaml

logger = logging.getLogger("rift")


def _load_ini_file(f: Path) -> configparser.ConfigParser:
    """
    Load an ini using configparser. Raise an exception if the file doesn't exist

    Args:
        f: Path to the file to load

    Returns: The ini file content

    """
    if not os.path.exists(f):
        raise Exception(f"ERROR ini file '{f}' doesn't exist")

    config = configparser.ConfigParser()
    config.optionxform = str  # to maintain upper case in config_reader file

    # read ini file
    config.read(f)
    return config


def read_ini(f: Path, keep_sections: bool = False) -> dict:
    """

    Args:
        f: Path to the ini file
        keep_sections: Whether to keep sections as keys or flatten the structure

    Returns:
        Dictionary with INI file contents
    """
    config = _load_ini_file(f)

    ini_dict = OrderedDict()
    # Iterate over sections in the INI file
    for section in config.sections():
        if keep_sections:
            ini_dict[section.upper()] = dict(config.items(section))
        else:
            # Flatten the section items into the main dictionary
            for key, value in config.items(section):
                ini_dict[key] = value

    # return to dict to raise an error in case of missing attributes
    return ini_dict


def read_json(f: Path):
    return json.loads(f.read_text())


def read_yaml(f: Path):
    try:
        with open(f) as file:
            data = yaml.load(file, Loader=yaml.SafeLoader)
        return data
    except FileNotFoundError:
        logger.error(f"File not found: {f}")
        raise
    except yaml.YAMLError as e:
        logger.error(f"Error parsing YAML file: {f} - {e}")
        raise



class ExecutionPathsMixin:
    def get_log_dir(self):
        return self.execution_path / "log"

    def get_work_dir(self):
        return self.execution_path / "work"

    def get_out_dir(self):
        return self.execution_path / "out"



from pathlib import Path

import rift.conf as settings
from rift.paths import utils


class TestCasePathsMixin:
    """
    Define the structure of directory necessary for a test case execution.
    No special rule about test cases path.
    """

    # current work dir is fundamental to reach the test implementation associated with a test case
    cwd = Path.cwd()

    @classmethod
    def get_source_path(cls):
        """Return source directory as the one which contains .git directory"""
        try:
            return utils.dot_git_parent_path(cls.cwd)
        except FileNotFoundError:
            return cls.cwd

    @classmethod
    def get_tests_base_path(cls) -> Path:
        """
        Tests base directory contains all test implementations. It is defined as:

        cwd/
        └── TESTS_BASE_DIR/

        Returns:
            Path where to find all test implementations
        """
        return cls.cwd / settings.TESTS_BASE_DIR

    @classmethod
    def get_test_implementation_path(cls, procedure_name: str) -> Path:
        """
        Test implementation contains the test entry point and all procedures
         necessary to run a test. It is defined as:

        cwd/
        └── TESTS_BASE_DIR/
            └── procedure_name/
        Args:
            procedure_name: name of the test implementation

        Returns:
            Path of the test implementation associated with the given test name
        """
        return cls.get_tests_base_path() / procedure_name

    @classmethod
    def get_test_entry_point(cls, procedure_name: str) -> Path:
        """
        The entry point is the executable called by rift starter to run the test
        cwd/
        └── TESTS_BASE_DIR/
            └── test_name/
                └── settings.TEST_ENTRY_POINT
        Args:
            procedure_name: name of the test implementation

        Returns:
            The entry point to run test_name
        """
        return cls.get_test_implementation_path(procedure_name) / settings.TEST_ENTRY_POINT



from pathlib import Path

from rift import conf as settings


class GlobalPathsMixin:
    root = Path(__file__).parent.parent.parent
    rift_package = root / "rift"

    shell_package = rift_package / "shell"
    bin_dir = rift_package / "bin"
    conf = rift_package / "conf"

    @classmethod
    def get_load_modules_script(cls):
        return cls.shell_package / settings.LOAD_MODULES

    @classmethod
    def get_starter(cls):
        return cls.shell_package / settings.RIFT_ENTRY_POINT

    @classmethod
    def get_shell_lib_entry_point(cls):
        return cls.shell_package / settings.SHELL_UTILS



import logging
from pathlib import Path

from rift import conf as settings
from rift.factories.test_case_factory import get_test_case
from rift.paths import PathManager

logger = logging.getLogger("rift")


def initialize_test_case(test_case_config: Path) -> None:
    """
    Initialize a test case if it doesn't exist yet'

    Args:
        test_case_config: Path where to initialize the test case
        execution_dir: Path where to execute the test case
    """
    if test_case_config.exists():
        logger.info(
            "Test case %s already exists", test_case_config.absolute().as_posix()
        )
        return
    test_name = test_case_config.stem

    ########################
    # TEST CASE
    ########################
    test_case_content = settings.MOCK_CONFIG
    test_case_content = test_case_content.replace(
        settings.TEST_NAME_PLACEHOLDER,
        test_name,
    )
    test_case_config.parent.mkdir(parents=True, exist_ok=True)
    test_case_config.write_text(test_case_content)

    ########################
    # TEST IMPLEMENTATION
    ########################
    test_case = get_test_case(test_case_config)
    path_manager = PathManager(test_case)
    main_path = path_manager.test_entry_point
    if main_path.exists():
        logger.info(f"Test implementation already exists: {main_path}")
        return

    main_path.parent.mkdir(parents=True, exist_ok=True)
    main_mock_content = settings.MOCK_ENTRY_POINT_CONTENT
    main_path.write_text(main_mock_content)



#!/usr/bin/env python

"""
This script provides an abstract interface to access the schedulers.
It provides a general access way also if the scheduler changes.
"""

__author__ = "Antonio Mariani"
__date__ = "2023/03/24"
__copyright__ = "CMCC Foundation - Euro-Mediterranean Center on Climate Change"

import subprocess
import os

LSF_SCHEDULER = "LSF"
SLURM_PARAMETER = "SLURM"

SUPPORTED_SCHEDULER = {LSF_SCHEDULER: ["bsub"], SLURM_PARAMETER: ["sbatch"]}

EMPTY_STRING = ""


def get_args():
    import argparse

    parse = argparse.ArgumentParser(description="Archiving tool")

    # General args

    parse.add_argument(
        "-c", "--cmd", action="store_true", help="Get scheduler command to submit jobs"
    )
    parse.add_argument(
        "-w", "--wait", action="store_true", help="To request waiting parameter"
    )
    parse.add_argument(
        "-ps",
        "--prios",
        action="store_true",
        help="Return a string with sequential prio_flag parameter-values",
    )
    parse.add_argument(
        "-pp",
        "--priop",
        action="store_true",
        help="Return a string with parallel prio_flag parameter-values",
    )
    parse.add_argument(
        "-i", "--id", action="store_true", help="Return a string with id info"
    )

    parse.add_argument(
        "-s",
        "--post_exec",
        type=str,
        default=None,
        help="Run command the job finishes - only LSF",
    )
    parse.add_argument(
        "-m",
        "--mem",
        type=str,
        default=None,
        help="To request memory request parameter",
    )
    parse.add_argument(
        "-t",
        "--run_time",
        type=str,
        default=None,
        help="To request jon running time parameter",
    )
    parse.add_argument(
        "-j", "--job_name", type=str, default=None, help="To request job name parameter"
    )

    parse.add_argument(
        "-e", "--err", type=str, default=None, help="Enable standard error log"
    )
    parse.add_argument(
        "-o", "--out", type=str, default=None, help="Enable standard output log"
    )
    parse.add_argument(
        "-l",
        "--log_dir",
        type=str,
        default=".",
        help="Change directory where store log files",
    )
    parse.add_argument("-q", "--queue", type=str, default=None, help="Set queue")
    parse.add_argument("-n", "--ntask", type=str, default=None, help="Number of task")

    return parse.parse_args()


def get_optional_attributes(key, dictionary: dict):
    value = dictionary.get(key, None)

    if value is not None and len(value) != 0:
        return value
    else:
        return None


def build_priority(app, sla):
    # sla and app must be present MANDATORY together
    if sla is not None and app is None:
        raise ValueError(f"ERROR sla is defined but app value is not defined")
    elif sla is None and app is not None:
        raise ValueError(f"ERROR sla is not defined but app value is defined")
    elif app is not None and sla is not None:
        priority_parameter = f"-sla {sla} -app {app}"
    else:
        priority_parameter = EMPTY_STRING
    return priority_parameter


class SchedulerResolver:
    def __init__(self):
        self._job_scheduler = None
        self._scheduler_cmd = None
        self._env: dict = os.environ

        self._load_scheduler_info()

    def _load_scheduler_info(self, verbose: bool = False):
        stdout = None if verbose else subprocess.PIPE
        stderr = None if verbose else subprocess.PIPE

        for scheduler, cmd_list in SUPPORTED_SCHEDULER.items():
            for cmd in cmd_list:
                result = subprocess.run(
                    [f"which {cmd} 2>/dev/null"],
                    stdout=stdout,
                    stderr=stderr,
                    text=True,
                    shell=True,
                )
                if result.returncode == 0:
                    if verbose:
                        print(f"Found scheduler: {scheduler}")
                    self._job_scheduler = scheduler
                    self._scheduler_cmd = cmd
                    break
        if self._job_scheduler is None:
            raise Exception("No scheduler has been found on the current machine")

    def get_cmd(self):
        return self._scheduler_cmd

    def get_post_exec_parameter(self, cmd_to_exec: str):
        if self._job_scheduler == SLURM_PARAMETER:
            return None
        elif self._job_scheduler == LSF_SCHEDULER:
            return f'-Ep "{cmd_to_exec}"'

    def get_wait_parameter(self):
        if self._job_scheduler == SLURM_PARAMETER:
            return "-W"
        elif self._job_scheduler == LSF_SCHEDULER:
            return "-K"

    def get_runtime_limit_parameter(self, r_limit):
        if self._job_scheduler == SLURM_PARAMETER:
            return f"-t {r_limit}"
        elif self._job_scheduler == LSF_SCHEDULER:
            return f"-W {r_limit}"

    def get_job_name_parameter(self, job_name):
        return f"-J {job_name}"

    def get_mem_request(self, memory_to_request):
        if self._job_scheduler == SLURM_PARAMETER:
            return f"--mem={memory_to_request}"
        elif self._job_scheduler == LSF_SCHEDULER:
            return f'-R "rusage[mem={memory_to_request}]"'

    def get_id(self):
        """LSF needs ID_PROJECT and SLURM needs ACCOUNT parameter, otherwise the job will be refused"""
        if self._job_scheduler == SLURM_PARAMETER:
            account_slurm = os.environ.get("MYE_ACCOUNT", None)
            if account_slurm is None:
                raise ValueError(
                    f"ERROR In {SLURM_PARAMETER} is mandatory the parameter MYE_ACCOUNT"
                )
            else:
                return f"-A {account_slurm}"
        elif self._job_scheduler == LSF_SCHEDULER:
            id_project = os.environ.get("MYE_LSF_Project", None)
            if id_project is None:
                raise ValueError(
                    f"ERROR In {LSF_SCHEDULER} is mandatory the parameter MYE_LSF_Project"
                )
            else:
                return f"-P {id_project}"

    def get_sequential_priority(self):
        if self._job_scheduler == SLURM_PARAMETER:
            priority_parameter = self.get_slurm_prio()
        elif self._job_scheduler == LSF_SCHEDULER:
            sla = get_optional_attributes("MYE_SC_serial", self._env)
            app = get_optional_attributes("MYE_LSF_SAPP", self._env)
            priority_parameter = build_priority(app, sla)
        else:
            self._unknown_scheduler()

        return priority_parameter

    def get_parallel_priority(self):
        if self._job_scheduler == SLURM_PARAMETER:
            priority_parameter = self.get_slurm_prio()
        elif self._job_scheduler == LSF_SCHEDULER:
            sla = get_optional_attributes("MYE_SC", self._env)
            app = get_optional_attributes("MYE_LSF_PAPP", self._env)
            priority_parameter = build_priority(app, sla)
        else:
            self._unknown_scheduler()

        return priority_parameter

    def get_slurm_prio(self):
        """In Slurm (G100) is used only parallel priority/queue"""
        qos = get_optional_attributes("MYE_QOS", self._env)
        reservation = get_optional_attributes("MYE_RESERVATION", self._env)
        priority_parameter = EMPTY_STRING
        if qos is not None:
            priority_parameter = f"-q {qos}"
        if reservation is not None:
            if priority_parameter is None:
                priority_parameter = f"--reservation {reservation}"
            else:
                priority_parameter = f"{priority_parameter} --reservation {reservation}"
        return priority_parameter

    def get_sequential_ntask(self):
        if self._job_scheduler == SLURM_PARAMETER:
            return f"--ntasks-per-node=1 -c 1"
        if self._job_scheduler == LSF_SCHEDULER:
            return f"-n 1"

    def get_queue(self, queue):
        if self._job_scheduler == SLURM_PARAMETER:
            return f"-p {queue}"
        if self._job_scheduler == LSF_SCHEDULER:
            return f"-q {queue}"

    def get_out_log(self, log_dir, log_name):
        return f"-o {log_dir}/{log_name}"

    def get_err_log(self, log_dir, log_name):
        return f"-e {log_dir}/{log_name}"

    def _unknown_scheduler(self):
        raise Exception(f"ERROR Unknown scheduler: {self._job_scheduler}")

    def get_parallel_ntask(self, ntask):
        if self._job_scheduler == SLURM_PARAMETER:
            raise Exception(f"ERROR Not supported mode for {self._job_scheduler}")
        if self._job_scheduler == LSF_SCHEDULER:
            return f"-n {ntask}"


def build_result(input_result, new_value):
    if len(input_result) == 0:
        return new_value
    elif new_value is not None and len(new_value) > 0:
        return f"{input_result} {new_value}"
    else:
        return input_result


def main():
    args = get_args()

    cmd = args.cmd
    wait = args.wait
    sequential_priority = args.prios
    parallel_priority = args.priop
    id_job = args.id

    post_exec = args.post_exec
    mem = args.mem
    run_time = args.run_time
    job_name = args.job_name

    err_log = args.err
    out_log = args.out
    log_dir = args.log_dir  # used only with err_log/out_log
    queue = args.queue
    ntask = args.ntask

    scheduler_resolver = SchedulerResolver()
    result = EMPTY_STRING

    # No value arguments
    if cmd:
        result = build_result(result, scheduler_resolver.get_cmd())
    if wait:
        result = build_result(result, scheduler_resolver.get_wait_parameter())
    if sequential_priority:
        result = build_result(result, scheduler_resolver.get_sequential_priority())
    if parallel_priority:
        result = build_result(result, scheduler_resolver.get_parallel_priority())
    if id_job:
        result = build_result(result, scheduler_resolver.get_id())

    # Value arguments
    if post_exec:
        result = build_result(
            result, scheduler_resolver.get_post_exec_parameter(post_exec)
        )
    if mem is not None:
        result = build_result(result, scheduler_resolver.get_mem_request(mem))
    if run_time:
        result = build_result(
            result, scheduler_resolver.get_runtime_limit_parameter(run_time)
        )
    if job_name:
        result = build_result(
            result, scheduler_resolver.get_job_name_parameter(job_name)
        )

    if err_log is not None:
        result = build_result(result, scheduler_resolver.get_err_log(log_dir, err_log))
    if err_log is not None:
        result = build_result(result, scheduler_resolver.get_out_log(log_dir, out_log))
    if queue is not None:
        result = build_result(result, scheduler_resolver.get_queue(queue))
    if ntask is not None:
        if ntask == 1:
            result = build_result(result, scheduler_resolver.get_sequential_ntask())
        else:
            result = build_result(result, scheduler_resolver.get_parallel_ntask(ntask))

    print(result)


if __name__ == "__main__":
    main()



import logging
from pathlib import Path
from typing import List

from pydantic import BaseModel, Field

from rift.model.exec_section import ExecSection
from rift.model.git_section import GitSection

logger = logging.getLogger("test_case")


class TestCase(BaseModel):
    path: Path
    environment: dict = Field(alias="ENVIRONMENT")
    exec: ExecSection = Field(alias="EXEC")
    repositories: List[GitSection] = Field(alias="REPOSITORIES")

    @property
    def name(self):
        return self.path.name

    @property
    def procedure_name(self):
        return self.exec.test_procedure

    def environ(self) -> dict:
        return {
            **self.environment,
            **self.exec.model_dump(by_alias=True)
        }



import logging.config

from rift import conf
from rift.utils.module_loading import import_string

DEFAULT_LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "require_debug_false": {
            "()": "rift.utils.log.RequireDebugFalse",
        },
        "require_debug_true": {
            "()": "rift.utils.log.RequireDebugTrue",
        },
    },
    "formatters": {
        "my_fmt": {
            "format": "[%(asctime)s] - %(levelname)s: %(message)s",
        }
    },
    "handlers": {
        "console": {
            "level": "INFO",
            "filters": ["require_debug_false"],
            "class": "logging.StreamHandler",
            "formatter": "my_fmt",
        },
        "console_with_debug": {
            "level": "DEBUG",
            "filters": ["require_debug_true"],
            "class": "logging.StreamHandler",
            "formatter": "my_fmt",
        },
    },
    "loggers": {
        "rift": {
            "handlers": ["console", "console_with_debug"],
            "level": "DEBUG",
        },
    },
}


class RequireDebugTrue(logging.Filter):
    def filter(self, record):
        return conf.DEBUG


class RequireDebugFalse(logging.Filter):
    def filter(self, record):
        return not conf.DEBUG


def configure_logging(logging_config, logging_settings):
    if logging_config:
        # First find the logging configuration function ...
        logging_config_func = import_string(logging_config)

        logging.config.dictConfig(DEFAULT_LOGGING)

        # ... then invoke it with the logging settings
        if logging_settings:
            logging_config_func(logging_settings)



import argparse
import importlib.metadata
import sys
from pathlib import Path

from rift import core


def get_args(raw_args=None) -> argparse.Namespace:
    parse = argparse.ArgumentParser(description="MedSea Test Automation Framework")
    # General args
    parse.add_argument("test_case_config", type=Path, help="Path of test case")
    parse.add_argument(
        "-v",
        dest="verbose",
        default=4,
        type=int,
        help="Verbose level from 1 (CRITICAL) to 5 (DEBUG). Default is 4 (INFO)",
    )
    parse.add_argument(
        "-w",
        dest="work_dir",
        type=Path,
        default=None,
        help="To change the test working dir. Default is ini_test_case_YYYYMMDDTHHmmss",
    )
    parse.add_argument(
        "-i",
        "--init",
        dest="to_initialize",
        default=False,
        action="store_true",
        help="If the test case doesn't exist, create it",
    )
    parse.add_argument(
        "-V",
        "--version",
        dest="get_version",
        default=False,
        action="store_true",
        help="Print version and exit",
    )

    if "-V" in sys.argv or "--version" in sys.argv:
        print(importlib.metadata.version("rift"))
        sys.exit(0)
    return parse.parse_args(raw_args)


if __name__ == "__main__":
    args: argparse.Namespace = get_args()
    core.execute(**vars(args))



#!/usr/bin/env python

import logging
import os
import subprocess
from pathlib import Path

logger = logging.getLogger("rift")


def evaluate_str_with_bash(input_str: str) -> str:
    """
    Retrieves the evaluated value of a Bash-style variable expression.

    Args:
        input_str (str): Input string to be evaluated and resolved.
            e.g., 'foo${var}' or 'foo${var:-"default"}' or $((expr)).
    Returns:
        str: The input str evaluated by the Bash shell.
    Raises:
        ValueError: If an unbound variable is encountered.

    Example:
        >>> evaluate_str_with_bash('foo${var}foo')
        'foobarfoo'
    """
    try:
        result = subprocess.run(
            f"set -o nounset; echo {input_str}",
            shell=True,
            check=True,
            capture_output=True,
            text=True,
        )
        return (
            result.stdout.strip()
        )  # Get the stdout output and strip any extra whitespace
    except subprocess.CalledProcessError as e:
        logger.error(f"Undefined variable: '{input_str}'")
        raise ValueError(f"Undefined variable: '{input_str}'") from e


def export_dict(environ: dict) -> None:
    """
    Export key-value pairs from a dictionary to environment variables.

    Args:
        environ (dict): Dictionary containing variables to export.
    """
    for key, value in environ.items():
        logger.debug(f"Try: export {key}: {value}")

        if value is None:
            logger.warning(f"Skipping variable {key}: {value}")
            continue

        value = str(value).lower() if isinstance(value, bool) else str(value)

        value_to_export = evaluate_str_with_bash(value) if "$" in value else value
        logger.info(f"Export {key} = {value_to_export}")
        os.environ[key] = value_to_export


def add_to_path(p: Path) -> None:
    """
    Args:
        p: Path to append to PATH
    """
    current_env_path = os.environ.get("PATH", "")
    export_dict({"PATH": f"{current_env_path}:{p.as_posix()}"})



from pathlib import Path


def dot_git_parent_path(p: Path) -> Path:
    """
    Find the parent directory of a git repository.
    Args:
        p: Starting point where start to scan

    Returns:
        Path of the directory that contains .git directory
    Raises:
        FileNotFoundError if / is reached without find any .git directory
    """
    current_p = p.resolve()
    while current_p.parent != current_p:
        if (current_p / ".git").is_dir():
            return current_p
        current_p = current_p.parent
    raise FileNotFoundError(
        f"fatal: not a git repository (or any parent up to mount point /) in {p}"
    )



# Mapping of file extensions to corresponding reader functions
import os
from pathlib import Path
from typing import Any, Dict, Callable, List

from rift.conf.global_settings import *  # noqa: F403
from rift.conf.mock_settings import *  # noqa: F403
from rift.conf.reader import read_ini, read_json, read_yaml

readers: Dict[Callable, List[str]] = {
    read_json: [".json", ".jsn"],
    read_yaml: [".yaml", ".yml"],
    read_ini: [".ini"],
}


def read_config(f: Path, *args, **kwargs) -> Any:
    """Reads config_reader file using the best strategy and returns dict"""
    # Normalize the file extension
    _, ext = os.path.splitext(f)
    ext = ext.lower()

    # Find the correct reader based on file extension
    for func, exts in readers.items():
        if ext in exts:
            return func(f, *args, **kwargs)

    # If no matching reader is found, raise an error
    raise ValueError(f"Unsupported file extension for file: {f}")



#!/usr/bin/env python

__author__ = "Antonio Mariani"
__email__ = "antonio.mariani@cmcc.it"
__version__ = "0.1.0"

import os
import re
import sys
import logging
from pathlib import Path

import xarray as xr
import numpy as np
import pandas as pd

from collections import namedtuple
from collections.abc import Sequence, Iterable
from typing import Optional, List

root_dir = str(Path(__file__).parent.parent.parent.resolve())
sys.path.append(root_dir)
from rift.lib import logging_config

# settings
logging_config.set_up()
logger = logging.getLogger("compare")
logger.info("ciao")

PASS = "PASSED"
FAIL = "FAILED"
DEFAULT_MAXDEPTH = 1  # negative value remove the limit on
DEFAULT_NAME_TO_COMPARE = ".nc"
DTYPE_NOT_CHECKED = ["S8", "S1", "O"]  # S8|S1:char, O: string
TIME_DTYPE = ["datetime64[ns]", "<M8[ns]"]

COMPARE_MESSAGE = namedtuple(
    "CompareMessage",
    [
        "result",
        "relative_error",
        "min_diff",
        "max_diff",
        "mask_equal",
        "file1",
        "file2",
        "variable",
        "description",
    ],
    defaults=["-" for _ in range(9)],
)


def get_args(raw_args=None):
    import argparse

    parse = argparse.ArgumentParser(description="netCDF Comparison Tool")
    # General args
    parse.add_argument("folder1", type=str, help="Path of first folder to compare")
    parse.add_argument("folder2", type=str, help="Path of second folder to compare")
    parse.add_argument(
        "--name",
        type=str,
        default=DEFAULT_NAME_TO_COMPARE,
        help="Name of the files to compare."
        "It can be a sub-set of the complete name or a regex expression",
    )
    parse.add_argument(
        "--maxdepth",
        type=int,
        default=DEFAULT_MAXDEPTH,
        help="Descend at most levels levels of directories below the "
        "starting-points. If set to -1 it scan all the subdirectories",
    )
    parse.add_argument(
        "--common_pattern",
        type=str,
        default=None,
        help="Common file pattern in two files to compare. "
        "Es mfsX_date.nc and expX_date.nc -> date.nc is the common part",
    )
    parse.add_argument(
        "--variables", nargs="+", default=None, help="Variable to compare"
    )
    parse.add_argument(
        "-v",
        dest="verbose_level",
        default=3,
        type=int,
        help="Verbose level from 1 (CRITICAL) to 5 (logger.debug). Default is 2 (ERROR)",
    )
    parse.add_argument(
        "--last_time_step",
        dest="last_time_step",
        action="store_true",
        default=False,
        help="If True, compare only the last time step available in each file",
    )

    return parse.parse_args(raw_args)


def all_match_are_satisfied(matching_strings: tuple, file2: str):
    if len(matching_strings) == 0:
        raise ValueError("Matching string list is empty")
    for match in matching_strings:
        if match not in file2:
            return False
        else:
            logger.debug(f"Found {match} in {file2}")

    return True


def get_match(pattern, string):
    if pattern[0] != "(" or pattern[-1] != ")":
        pattern = f"({pattern})"  # force to use group refex as search
    match_object = re.search(pattern, string)
    if match_object is not None:
        return match_object.groups()
    else:
        return None


def get_file_list_to_compare_with_match(
    sequence1: Sequence, sequence2, match_pattern: str
):
    """
    Given a regex match pattern, it finds a fileX in sequence1 that match the regex expression,
        then try to find in sequence2 a fileY that match the regex pattern with the same value of fileX.
        Example:
            match_pattern = \d{7}_\d{8}
            fileX = MO_PR_PF_*3901977_20221201*_p01_20221201_20221201.nc
            fileY = GL_LATEST_PR_PF_*3901977_20221201*.nc
        Between ** character has been highlighted the pattern that match with regex match_patten
    Args:
        sequence1: An iterable where find the files that match with match_pattern
        sequence2: An iterable where find the files that match match_pattern with the same values of the files
            from sequence1
        match_pattern: A regex expression

    Returns: A list of tuple. Each tuple contains two files that have the same value in correspondence of match_pattern
    """
    not_found = 0
    match_list = list()
    for file in sequence1:
        matching_strings = get_match(match_pattern, file)
        if matching_strings:
            match_found = False
            for file2 in sequence2:
                filename2 = os.path.basename(file2)
                if all_match_are_satisfied(matching_strings, filename2):
                    match_found = True
                    logger.debug(f"Found two files that match")
                    logger.debug(f"\t- {file}")
                    logger.debug(f"\t- {file2}")
                    match_list.append((file, file2))
            if not match_found:
                print(f"{FAIL} No matching found for {file}")
                not_found += 1
        else:
            logger.debug(f"The file {file} doesn't match with {match_pattern}")

    return match_list, not_found


def walklevel(input_dir, level=1):
    """Generator function, usage: for root, dirs, files in walklevel: # do something"""
    input_dir = input_dir.rstrip(os.path.sep)
    assert os.path.isdir(input_dir)
    num_sep = input_dir.count(os.path.sep)
    max_depth_level = (
        num_sep + level - 1
    )  # -1 used to have the same behavior of find bash command
    for root, dirs, files in os.walk(input_dir):
        yield root, dirs, files
        num_sep_this = root.count(os.path.sep)
        # level < 0 is a special case used to explore all the subdirs
        if num_sep_this >= max_depth_level and level > 0:
            del dirs[:]


def get_file_list_to_compare(nc_file_list1: list, nc_file_list2: list):
    not_found = 0
    file_list_to_compare = list()
    for file1 in nc_file_list1:
        filename1 = os.path.basename(file1)
        match_file2 = [file2 for file2 in nc_file_list2 if filename1 in file2]
        if len(match_file2) == 0:
            print(f"{FAIL} No matching found for {file1}")
            not_found += 1
        else:
            file_list_to_compare += [(file1, file2) for file2 in match_file2]

    return file_list_to_compare, not_found


def safe_open_dataset(input_file: str) -> (Optional[xr.Dataset], str):
    try:
        return xr.open_dataset(input_file), None
    except FileNotFoundError:
        err_msg = "File not found or path is incorrect."
    except OSError as e:
        err_msg = f"An OS error occurred: {e}"
    except ValueError as e:
        err_msg = f"Value error occurred: {e}"
    except RuntimeError as e:
        err_msg = f"Runtime error occurred: {e}"
    except KeyError as e:
        err_msg = f"Key error occurred: {e}"
    except IndexError as e:
        err_msg = f"Index error occurred: {e}"
    except Exception as e:
        err_msg = f"An unexpected error occurred: {e}"

    # in case of exception
    return None, err_msg


def find_time_dims_name(dims: Iterable) -> str:
    time_dims_name = [dim for dim in dims if "time" in dim]
    if len(time_dims_name) == 0:
        return None
    if len(time_dims_name) > 1:
        raise ValueError(
            f"Found more than 1 time dimension: {', '.join(time_dims_name)}"
        )
    return time_dims_name.pop()


def compare_datasets(
    file1, file2, variables_to_compare: list, last_time_step: bool
) -> List:
    logger.info(f"Comparing {file1} with {file2}")
    dataset1, err_msg = safe_open_dataset(file1)
    if err_msg is not None:
        return [COMPARE_MESSAGE(description=err_msg)]
    dataset2, err_msg = safe_open_dataset(file2)
    if err_msg is not None:
        return [COMPARE_MESSAGE(description=err_msg)]

    # keep only float vars
    if variables_to_compare:
        dataset1_vars_list = variables_to_compare
    else:
        dataset1_vars_list, err_msg = get_dataset_variables(dataset1)

    if err_msg is not None:
        # an error message has been already printed at this point
        return [COMPARE_MESSAGE(description=err_msg)]

    logger.debug(f"Variables to check: {dataset1_vars_list}")

    result = []
    for var in dataset1_vars_list:
        logger.info(f"Checking {var}")

        field1 = dataset1[var]

        # missing variables in comparison file
        try:
            field2 = dataset2[var]
        except Exception as e:
            result.append(
                COMPARE_MESSAGE(description=f"cannot read {var} from {file2}: {e}")
            )
            continue

        # if last_time_step option is used:
        # - drop all time steps except last one
        # - do not compare time variables
        if last_time_step:
            time_dims_name = find_time_dims_name(field1.dims)
            if time_dims_name and field1.shape[0] > 1:
                field1 = field1.drop_isel(
                    {time_dims_name: [t for t in range(field1.shape[0] - 1)]}
                )
            if time_dims_name and field2.shape[0] > 1:
                field2 = field2.drop_isel(
                    {time_dims_name: [t for t in range(field2.shape[0] - 1)]}
                )
            if "time" in var:
                continue

        # dimensions mismatch
        if field1.shape != field2.shape:
            result.append(
                COMPARE_MESSAGE(
                    description=f"Can't compare {var} in {file1} and in {file2} with shapes {field1.shape} {field2.shape}"
                )
            )
            continue

        array1, mask_array1 = np.array(field1.values), field1.to_masked_array()
        array2, mask_array2 = np.array(field2.values), field2.to_masked_array()

        # try computing difference
        try:
            difference_field: np.ma.MaskedArray = mask_array1 - mask_array2
        except Exception as e:
            result.append(
                COMPARE_MESSAGE(
                    description=f"an unknown error occurs while comparing {var}: {e}"
                )
            )
            continue

        # get statistics
        max_difference = float(difference_field.max())
        min_difference = float(difference_field.min())
        mask_is_equal = np.array_equal(mask_array1.mask, mask_array2.mask)

        if min_difference is np.nan and max_difference is np.nan:
            min_difference = 0
            max_difference = 0
            descr = "WARNING all nan values found - comparison is not possible"
            rel_err = 0
        else:
            rel_err = compute_relative_error(difference_field, field2)
            descr = "-"

        check_result = FAIL
        if (
            min_difference == 0
            and max_difference == 0
            and mask_is_equal
            and rel_err == 0
        ):
            check_result = PASS

        result.append(
            COMPARE_MESSAGE(
                result=check_result,
                relative_error=f"{rel_err:.2e}",
                min_diff=f"{min_difference:.2e}",
                max_diff=f"{max_difference:.2e}",
                mask_equal=f"{mask_is_equal}",
                file1=f"{os.path.basename(file1)}",
                file2=f"{os.path.basename(file2)}",
                variable=f"{var}",
                description=descr,
            )
        )
    return result


def compute_relative_error(diff: np.ma.MaskedArray, field2: xr.DataArray):
    diff_no_nan = np.nan_to_num(diff, nan=0)
    if np.all(diff_no_nan == 00):
        return 0.0

    if field2.dtype in TIME_DTYPE:
        field2_values = field2.values.view("int64")
    else:
        field2_values = field2.values

    array2_no_nan = np.nan_to_num(field2_values, nan=0)
    try:
        # Suppress division by zero and invalid value warnings
        with np.errstate(divide="ignore", invalid="ignore"):
            array2_abs = np.abs(array2_no_nan)
            rel_err = np.max(diff_no_nan / array2_abs)
    except Exception as e:
        logger.warning(f"An error occurred when computing relative error: {e}")
        rel_err = np.nan

    if field2.dtype in TIME_DTYPE:
        return rel_err / np.timedelta64(1, "s")
    return rel_err


def get_dataset_variables(dataset: xr.Dataset):
    """Extract all non char/str variables included dimension from a dataset"""
    variables = []
    try:
        variables.extend(
            [
                var_name
                for var_name in dataset.data_vars
                if dataset[var_name].dtype not in DTYPE_NOT_CHECKED
            ]
        )
    except Exception as e:
        return None, f"Cannot extract variables from {dataset}: {e}"

    try:
        variables.extend(
            [
                var_name
                for var_name in dataset.dims
                if dataset[var_name].dtype not in DTYPE_NOT_CHECKED
            ]
        )
    except Exception as e:
        return None, f"Cannot extract dimensions from {dataset}: {e}"

    return variables, None


def main(raw_args=None):
    args = get_args(raw_args)
    folder1 = args.folder1
    folder2 = args.folder2
    filter_name = args.name
    common_pattern = args.common_pattern
    maxdepth = args.maxdepth
    variables_to_compare: list = args.variables
    verbose_level: int = args.verbose_level
    last_time_step = args.last_time_step

    # set verbosity
    logging_config.set_level(verbose_level)

    # read input file list
    nc_file_list1 = load_file_list(filter_name, folder1, maxdepth)
    nc_file_list2 = load_file_list(filter_name, folder2, maxdepth)

    # filter file list to compare
    if common_pattern is None:
        files_to_compare, not_found = get_file_list_to_compare(
            nc_file_list1, nc_file_list2
        )
    else:
        files_to_compare, not_found = get_file_list_to_compare_with_match(
            nc_file_list1, nc_file_list2, common_pattern
        )

    # start comparison
    results = []
    errors_found = not_found
    for file1, file2 in files_to_compare:
        df = pd.DataFrame(
            [],
            columns=[
                "Result",
                "Relative error",
                "Min Diff",
                "Max Diff",
                "Mask Equal",
                "Reference File",
                "Comparison File",
                "Variable",
                "Description",
            ],
        )
        result = compare_datasets(file1, file2, variables_to_compare, last_time_step)
        for row_data in result:
            df.loc[len(df)] = list(row_data)
        df_to_print = df.drop(["Comparison File", "Reference File"], axis=1)
        print(f"\n- Reference file: {file1}")
        print(f"- Comparison file: {file2}")
        print(df_to_print.to_string(index=False))
        if (df["Result"] == "FAILED").any():
            errors_found += 1
        results.append(df)

    if errors_found > 0:
        exit(1)


def load_file_list(filter_name, folder1, maxdepth):
    nc_file_list1 = [
        os.path.join(root, f)
        for root, dirs, files in walklevel(folder1, maxdepth)
        for f in files
        if get_match(filter_name, os.path.join(root, f))
    ]
    return nc_file_list1


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import logging
import os
import subprocess

import psutil

logger = logging.getLogger("rift")


class Scheduler:
    def __init__(self, start_script, main_procedure, load_module_script, log_dir):
        """
        :param start_script: Start test framework script
        :param main_procedure: Main script of test procedure
        :param load_module_script: Script used to load modules
        :param log_dir: Log directory
        """

        self.start_script = start_script
        self.main_procedure = main_procedure
        self.load_module_script = load_module_script
        self.log_dir = log_dir

        self._result = None

    def exec(self):
        stdout = None  # subprocess.PIPE to block stdout
        stderr = None  # subprocess.PIPE to block stderr

        logger.info(f"Starter: {self.start_script}   # added automatically")
        logger.info(f"Main: {self.main_procedure}")
        logger.info(f"Load modules: {self.load_module_script}")

        try:
            logger.info("EXECUTION START")

            print(
                f"bash {self.start_script} {self.main_procedure} {self.load_module_script}"
            )
            self._result = subprocess.run(
                [
                    f"bash {self.start_script} {self.main_procedure} {self.load_module_script}"
                ],
                stdout=stdout,
                stderr=stderr,
                text=True,
                shell=True,
            )
            logger.info("EXECUTION_DONE")
        except KeyboardInterrupt:
            logger.critical("Caught Ctrl + C, starting exit procedure...")
            pid_file = self.log_dir + "/pid"
            if os.path.exists(pid_file):  # to kill father process and all kid processes
                with open(self.log_dir + "/pid") as f:
                    pid = f.read().strip("\n")
                    kill_process_with_child(int(pid))
                logger.info("EXECUTION_DONE")
                exit(1)
            else:
                logger.error(f"Cannot find pid file: {pid_file}")

    def get_return_code(self):
        return self._result.returncode


def kill_process_with_child(root_pid):
    try:
        logger.debug(f"Root process PID: {root_pid}")
        root_process = psutil.Process(root_pid)
        logger.debug(f"Root process name: {root_process.name()}")

    except Exception as e:
        logger.debug(f"{e}")
        logger.warning(
            f"Root process with PID: {root_pid} is already terminated or does not exist, exiting..."
        )
        return

    for child in root_process.children(recursive=True):
        safe_kill(child.pid)

    logger.info("Terminating root process")
    safe_kill(root_process.pid)

    logger.info("Process termination completed successfully.")


def safe_kill(pid, timeout=5):
    try:
        logger.info(f"Kill {pid}")
        process = psutil.Process(pid)
        process.kill()
        process.wait(timeout=timeout)
    except psutil.NoSuchProcess:
        logger.debug(f"Process with PID {pid} not found.")
    except psutil.AccessDenied:  # this should never happen
        logger.error(f"Access denied when trying to terminate process with PID {pid}.")
    except psutil.TimeoutExpired:
        logger.warning(
            f"Timeout expired while waiting for process with PID {pid} to terminate."
        )
    except Exception as e:
        logger.error(f"Error occurred while terminating process with PID {pid}: {e}")



########################
# TEST STRUCTURE
########################

# where configuration file are expected to be put
TEST_CASES_BASE_DIR = "test_cases"
# where expect to find the test implementation
TESTS_BASE_DIR = "tests"

# Directory names
REFERENCE_TEST_DIR = "CMCC_local"  # used to redirect procedure definition and test execution in this directory
BASEDIR_PROCEDURES = (
    "tests"  # in CMCC_local: directory where to put the test procedure directories
)
SHELL_LIB_DIRNAME = "shell"  # contains shell libraries imported by starter
PYTHON_LIB_DIRNAME = "lib"  # contains python libraries
BIN_DIRNAME = "bin"


# Script names
LOAD_MODULES = "load_modules.sh"  # used to load the modules declared in the test case
TEST_ENTRY_POINT = "main.sh"  # the starting point of the test
RIFT_ENTRY_POINT = "start.sh"  # the first script called by the rift
SHELL_UTILS = "utils.sh"  # the first script called by the manager

########################
# MOCK
########################
MOCK_PROCEDURE_NAME = "mock_test_procedure"
MOCK_MAIN = "main.sh"
MOCK_TEST_CASE = "test_case.ini"

########################
# LOG
########################

# The callable to use to configure logging
LOGGING_CONFIG = "logging.config.dictConfig"

# Custom logging configuration.
LOGGING = {}

DEBUG = False



TEST_NAME_PLACEHOLDER = '{{TEST_PROCEDURE_NAME}}'

MOCK_CONFIG = """
[Environment]

MY_VAR = This is a test
MY_VAR2 = Another custom variable


[Exec]
test_procedure = {0}
; write as module1 module2 ... moduleX -> automatic loaded before to call the test procedure
modules =
; if defined, the conda environment is automatic loaded before to call the test procedure
conda_env =
    """.format(TEST_NAME_PLACEHOLDER)

MOCK_ENTRY_POINT_CONTENT = """
#! /usr/bin/env bash

#
# Author: Name Surname (mail@cmcc.it)
#
#% Common template to start a bash development
#


########################
# BASH SETTINGS
########################
set -o errexit  # abort on nonzero exitstatus
set -o nounset  # abort on unbound variable
set -o pipefail # don't hide errors within pipes

########################
# VARIABLES
########################
script_name=$(basename "${0}")
script_dir=$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )

work_dir="${WORK_DIR}"
my_var="${MY_VAR}"
readonly script_name script_dir work_dir

########################
# MAIN
########################
main() {
    cd "${work_dir}"    # where a task must work

    echo "Hello World! from ${script_dir}/${script_name}"
    echo "Variable declared into test_case: ${my_var}"
}

main "${@}"

    """



#!/usr/bin/env python

import os
import subprocess
import logging
from pathlib import Path

logger = logging.getLogger("rift")


def clone(url, branch, dst) -> None:
    """
    Clone a git repo with designed branch to dst
    :param url: SSH url or local path from where to clone a repo
    :param branch: Git branch name
    :param dst: Where to clone the repo
    """
    if os.path.exists(dst):
        logger.warning(f"{dst} already exists")
        return

    if is_local_repo(url):
        link_local_repo(src=url, dst=dst)
    elif is_github_url(url):
        download_from_github(url, branch, dst)
    else:
        raise ValueError(f"Invalid url: {url}")


def is_local_repo(path: str) -> bool:
    """Return True if path is a clone of a git repository"""
    return os.path.exists(os.path.join(path, ".git"))


def is_github_url(url: str) -> bool:
    """Return True if url is an ssh git url"""
    return url.startswith("git@")


def download_from_github(url: str, branch: str, dst: Path) -> None:
    """
    Args:
        url: git url
        branch: branch to clone
        dst: Path where to clone the repo

    Raises:
        FileExistsError if dst already exists
    """
    dst.mkdir()

    try:
        subprocess.run(
            ["git", "clone", "-b", branch, url, dst],
            stdout=None,
            stderr=None,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        logging.exception(f"An error occurred while downloading from GIT: {e}")
        raise e


def link_local_repo(src, dst):
    logger.info(f"Cloning from {src} to {dst}")
    os.symlink(src, dst)



#!/usr/bin/env python
import logging
import os.path
from pathlib import Path
from typing import List

from rift.conf import environment
from rift.core import mock
from rift.core import source
from rift.factories.test_case_factory import get_test_case
from rift.core import scheduler
from rift.paths import PathManager

logger = logging.getLogger("rift")


def execute(test_case_config: Path, work_dir: Path, to_initialize: bool):
    test_case_config = test_case_config.resolve()
    ########################
    # INIT MOCK (if needed)
    ########################
    if to_initialize:
        logger.info(f"Init test case: {test_case_config}")
        mock.initialize_test_case(test_case_config)

    ########################
    # LOAD TEST CASE - PATH MANAGER
    ########################
    logger.info(f"Loading test case configuration: {test_case_config}")
    test_case = get_test_case(test_case_config)
    path_manager = PathManager(test_case, work_dir)

    ########################
    # PREPARE ENVIRONMENT
    ########################

    # append to PATH
    paths_to_export = path_manager.get_paths_to_append_to_env_path()
    for p in paths_to_export:
        environment.add_to_path(p)

    # export variables
    environ_to_export = test_case.environ()
    test_path_structure = path_manager.get_test_execution_structure()

    environment.export_dict(environ_to_export)
    environment.export_dict(test_path_structure)

    ########################
    # CLONE REPOS
    ########################
    for repo in test_case.repositories:
        source.clone(repo.url, repo.branch, dst=path_manager.external_repos_base_path)

    ########################
    # EXECUTION
    ########################
    schd = scheduler.Scheduler(
        path_manager.starter,
        path_manager.test_entry_point,
        path_manager.module_loads_entry_point,
        path_manager.log_dir,
    )
    schd.exec()
    exit_code = schd.get_return_code()
    if exit_code != 0:
        logger.error(f"TEST FAILS - exit code: {exit_code}")
        exit(exit_code)


if __name__ == "__main__":
    execute()



from datetime import datetime
from pathlib import Path
from typing import Dict

from rift.model import TestCase
from rift.paths.execution_paths_mixin import ExecutionPathsMixin
from rift.paths.global_paths_mixin import GlobalPathsMixin
from rift.paths.test_case_paths_mixin import TestCasePathsMixin


def timestamp():
    return datetime.now().strftime("%Y%m%dT%H%M%S")


class PathManager(GlobalPathsMixin, ExecutionPathsMixin, TestCasePathsMixin):
    def __init__(self, test_case: TestCase, execution_path: Path = None):
        self._test_case = test_case
        if execution_path is not None:
            self._execution_path = execution_path
        else:
            self._execution_path = self.cwd / f"{self._test_case.name}_{timestamp()}"

    @property
    def test_case(self) -> TestCase:
        return self._test_case

    @property
    def execution_path(self) -> Path:
        return self._execution_path

    @property
    def implementation_path(self) -> Path:
        return self.get_test_implementation_path(
            procedure_name=self.test_case.procedure_name
        )

    @property
    def shell_library(self) -> Path:
        return self.get_shell_lib_entry_point()

    @property
    def external_repos_base_path(self) -> Path:
        return self.cwd

    @property
    def starter(self) -> Path:
        return self.get_starter()

    @property
    def test_entry_point(self) -> Path:
        return self.get_test_entry_point(self.test_case.procedure_name)

    @property
    def module_loads_entry_point(self) -> Path:
        return self.get_load_modules_script()

    @property
    def source_dir(self) -> Path:
        return self.get_source_path()

    @property
    def log_dir(self) -> Path:
        return self.get_log_dir()

    @property
    def work_dir(self) -> Path:
        return self.get_work_dir()

    @property
    def out_dir(self) -> Path:
        return self.get_out_dir()

    def get_test_execution_structure(self) -> Dict[str, Path]:
        return {
            # paths related to rift and user cwd
            "SHELL_LIB": self.shell_library,
            "TEST_DIR": self.implementation_path,
            # paths related to specific test execution
            "SOURCE_DIR": self.get_source_path(),
            "LOG_DIR": self.get_log_dir(),
            "WORK_DIR": self.get_work_dir(),
            "OUT_DIR": self.get_out_dir(),
        }

    def get_paths_to_append_to_env_path(self):
        return [
            self.bin_dir,
            self.implementation_path,
        ]



from pathlib import Path

import rift.conf as settings
from rift.model import TestCase


def get_test_case(test_case_config: Path) -> TestCase:
    test_case_content: dict = settings.read_config(test_case_config.resolve(), keep_sections=True)
    git_keys_section = [k for k in test_case_content.keys() if k.startswith('GIT_')]

    # default value if no repos in test_case_config
    git_sections = []
    for k in git_keys_section:
        git_sections.append(test_case_content.pop(k))

    test_case_content['REPOSITORIES'] = git_sections
    return TestCase(path=test_case_config, **test_case_content)




import rift
from rift.core import main
from rift.core import scheduler


def execute(**kwargs):
    rift.setup()
    kwargs.pop("verbose")
    get_version = kwargs.pop("get_version")

    if get_version:
        exit(0)

    main.execute(**kwargs)
